#
# TrackerUtils.py
# 
# Peter Bengtsson <mail@peterbe.com>
# Utility scripts for SiteTrackerProduct
#
# Many actions on objects here are unrestricted,
# so that's why I have to manually check these things
# sometimes.
# If you find a way to hack this product (as an instance), email me :)
#

# python 
import string

# zope
from DateTime import DateTime
from DocumentTemplate import sequence
from DocumentTemplate.DT_Var import structured_text
from Products.FSSession import FSSession

  
def hasStoredFilter(self):
    """
        Return true if cookieobject has a xxx_filteroptions
        variable stored.
    """
    cookieobject = self.cookieobject()
    trackerroot = getRoot(self)

    tfc = "%s_filteroptions"%trackerroot.id
    
    if cookieobject.has_key(tfc):
        return 1
    else:
        return 0

def trytofindSessionObject(self,trackerid):
    tracker = getattr(self,trackerid)
    fssessions = tracker.aq_parent.superValues('FSSession')
    if len(fssessions) > 1:
        cookie = {'cookieobjectid':fssessions[0].id}
        tracker.propertysheets.properties.manage_changeProperties(cookie)
    return
    
def FillListingRequest(self):
    """
        Prepopulate REQUEST
        with saved filteroptions from cookie
        Don't use this if the url is fiddled with manually,
        like in url?type=bug
    """
    req = self.REQUEST
    cookieobject = self.cookieobject()
    trackerroot = getRoot(self)

    #TrackerFilteroptionsCookiename
    tfc = "%s_filteroptions"%trackerroot.id
    
    usecookie=1
    # things expected to be set manually in URL
    reasons_toignore= ['section','type','urgency','status','fromname','email']
    for reason in reasons_toignore:
        if reason in req.form.keys():
            usecookie = 0

    if usecookie and req.has_key('reset_filter') and cookieobject.has_key(tfc):
        # remove cookie and move on
        cookieobject.delete(tfc)

    if usecookie and cookieobject.has_key(tfc) and not req.has_key('filteroptions'):
        for key,value in cookieobject[tfc].items():
            req.set(key,value)
            
    return "Filled Listing Request"

def FillRequest(self):
    """
    Loop though a list of variables
    if it doesn't exist in either self
    or REQUEST, fill them into REQUEST
    as '' values
    """
    req = self.REQUEST

    propertiesNulls 	= ['subject','fromname','email','description','url2issue','urgency','type']
    propertiesZero 	= ['confidential','hide_me']
    propertiesEmpty 	= ['sections']
    
    nullified = []
    for property in propertiesNulls:
        if not req.has_key(property):
            req.set(property,'')
            
    for property in propertiesZero:
        if not req.has_key(property):
            req.set(property,0)
            
    for property in propertiesEmpty:
        if not req.has_key(property):
            req.set(property,[])
    
    # Also, for all list variables, make sure that only
    # one item doesn't make it a string
    for property in propertiesEmpty:
        if type(req[property]) == type(''):
            req[property] = [req[property]]
            
    # properties that have default values
    propertiesDefault	= [['status','open']]
    for property in propertiesDefault:
        if not req.has_key(property[0]):
            req.set(property[0],property[1])

    # special one
    try:
        if req['url2issue'][0] != '/' and req['url2issue'][:4] !='http':
            req.set('url2issue','http://%s'%req['url2issue'])
    except:pass # no big deal


    return ""
    

def ShowDescription(text):
    """
    Display text, using harmless HTML
    """

    st=text

    st = st.replace('<','&lt;')
    st = st.replace('>','&gt;')

    st = structured_text(st)
    
    # BUG in structured_text in Zope 2.4.0
    # it returns these annoying tags.
    st = st.replace('<html>','')
    st = st.replace('<body>','')
    st = st.replace('</body>','')
    st = st.replace('</html>','')

    return st
    

def getNowdate():
    """
    This method determines the format 
    with which new objects get a date property
    """
    return DateTime()
    
def SubmitIssue(self):
    """
    1) Check data
    2) Try to create issue
    	2a) If success, RESPONSE.redirect to issue plus Thank you message
    	2b) If failure, print failed data and urge to submit again
    """
    req = self.REQUEST
    
    randomID_int_length = 5
    if hasattr(self, 'randomID_int_length'):
        # use the Zope version instead
        randomID_int_length = self.randomID_int_length

    failures = {}
    
    # ! Change correspondingly to you pref cookie object !
    cookieobject = self.cookieobject()

    
 
    #
    # Tune data
    #
    
    # strip whitespace
    for property in ['subject','fromname','email','url2issue']:
        req[property] = req[property].strip()

    # quick add doesn't necessarily ask for email,
    # when it alread exists in the cookie. (lets do it for name too)
    if req.email=='' and cookieobject.has_key(self.email_cookiekey):
        req.set('email',cookieobject[self.email_cookiekey])
    if req.fromname=='' and cookieobject.has_key(self.name_cookiekey):
        req.set('fromname',cookieobject[self.name_cookiekey])
        
        
        
    # set sections to default values if empty
    if len(req.sections) ==0:
        req.sections = self.defaultsections
    
    #
    # Check data
    #
    
    # subject mustnt be empty
    if req.subject =='':
        # no input
        failures['Subject'] = "You must at least enter a subject"
        
    # to prevent problems with sending mail
    if not ValidEmailAddress(req.email):
        failures['Invalid Emailaddress'] = "To be able to notify you about the Issue, you must enter a valid email address."
        # hack. Leave a trace in the REQUEST
        # for the QuickAddIssue
        req.set('invalidemailaddress',1)
    
    if len(failures) >0:
        return self.ExplainSubmitErrors(failures)
        
    else:
        #
        # Let's Add it!!
        #
        
        # 1) Generate Random Issue Id
        genid = RandomID(self, "Issue", 5)
        
        # 3) Set cookie of name and email using cookieobject
        if req.has_key('fromname') and req.fromname <> '':
            cookieobject.set(self.name_cookiekey, req.fromname)
        if req.has_key('email') and req.email <> '':
            cookieobject.set(self.email_cookiekey, req.email)
            
        # 4) Try to add Issue
        # add Issue
        self.SiteTrackerIssue.createInObjectManager(genid,self.REQUEST)
                
        # get object from id
        issueobject = getattr(self, genid)

        # prepopulate empty REQUEST data
        FillRequest(self)

        # filter out empty item from sections
        sections_newlist=[]
        for section in req.sections:
            if section!='':
                sections_newlist.append(section)
                
        # add all the properties
        props = {}
        props['title'   ]       = req.subject
        props['fromname']	= req.fromname
        props['email'   ]	= req.email
        props['url2issue'] 	= req.url2issue
        props['type'    ] 	= req.type
        props['urgency' ] 	= req.urgency
        props['description'] 	= req.description
        props['confidential'] 	= int(req.confidential)
        props['hide_me' ]	= int(req.hide_me)
        props['issuedate'] 	= getNowdate()
        props['status'  ]	= req.status
        props['sections']       = sections_newlist

        # update issue object
        issueobject.propertysheets.properties.manage_changeProperties(props)
        
        # Catalog yourself
        self.catalog_object(issueobject)

        # 5) Redirect
        redirect_url = '%s'%(issueobject.absolute_url())
        req.RESPONSE.redirect(redirect_url)
        
        # Still here?
        # Hmm show a simple URL
        # Remember that Zope expects this method to return
        # a list of failures.
        return ['<h2><a href="%s">Issue Added</h2>'%redirect_url]
        

def RandomID(catalog, start, length, meta_type='Site Tracker Issue',increment=None):
    # digits
    if increment is None:
        idnr=len(catalog.objectIds(meta_type))+1
        increment = idnr+1
    else:
        increment=increment+1
        idnr=increment
        
    r=string.zfill(str(idnr),length)
    id = "%s-%s"%(start, r)
    if hasattr(catalog, id):
        # ah! Id already exists, try again
        return RandomID(catalog, start, length,meta_type,increment=increment)
    else:
        return id


def ValidEmailAddress(s):
    """
    NEEDS WORK!!!
    Have to check if a string is a valid email address    
    """
    if s == '' or s.find(' ') > -1 or len(s.split('@')) != 2:
        return 0
    elif s.split('@')[1].count('.') < 1:
        return 0
    return 1
    

def getRoot(self):
    """ superValues() does what objectValues() does, but
    continues down the path until found. If two Site Tracker
    instances are in the same folder, they are picked out by
    id unsmartly, hence this little test"""
    supervalues = self.superValues('Site Tracker')
    for supervalue in supervalues:
        if supervalue.id in self.REQUEST.URL1.split('/'):
            return supervalue
    else:
        return supervalues[0]

def ListIssueFiltered(self):
    if self.REQUEST.has_key('q') and self.REQUEST['q'] != '':
        q=self.REQUEST['q']
        catalogs = self({'title':q}) + \
                   self({'description':q}) + \
                  self({'fromname':q}) + \
                  self({'email':q})
        seq=[]
        for brain in catalogs:
            try:
                issueobject = getattr(self,brain.id)
                if issueobject not in seq:
                    seq.append(issueobject)
            except:
                # probably because issue has been deleted manually
                pass
    else:
        seq = self.objectValues('Site Tracker Issue')
    return _ListIssueFiltered(self,seq)

def _ListIssueFiltered(self,objsequence):
    rlist = []
    req = self.REQUEST

    cookieobject = self.cookieobject()

    try:
        reverse=int(req.reverse)
    except:
        reverse=0

    trackerroot = getRoot(self)

    #TrackerFilteroptionsCookiename
    tfc = "%s_filteroptions"%trackerroot.id
    
    # remember filteroptions
    if req.has_key('filteroptions'):
        tracker_filteroptions={}
        for each in ['showtypes','showsections','showstatuses','showfromname','showurgencies','showemail']:
            try:
                tracker_filteroptions[each]=req[each]
            except:pass
        cookieobject.set(tfc,tracker_filteroptions)
        
   
    #
    # prepare a master filter
    #
    
    filter ={}
    
    # statuses
    if req.has_key('showstatuses'):
        filter['status'] = req.showstatuses
    elif req.has_key('status'):
        filter['status'] = [req.status]
    else:
        filter['status'] = self.statuses
        
    # sections
    if req.has_key('showsections'):
        filter['sections'] = req.showsections
    elif req.has_key('section'):
        filter['sections'] = [req.section]
    else:
        filter['sections'] = self.sections_options

    # urgencies
    if req.has_key('showurgencies'):
        filter['urgency'] = req.showurgencies
    elif req.has_key('urgency'):
        filter['urgency'] = [req.urgency]
    else:
        filter['urgency'] = self.urgencies
        
    # types
    if req.has_key('showtypes'):
        filter['type'] = req.showtypes
    elif req.has_key('type'):
        filter['type'] = [req.type]
    else:
        filter['type'] = self.types
        
    # fromname
    if req.has_key('showfromname') and \
       req.showfromname <> self.when_ignore_word:
            filter['fromname'] = string.strip(req.showfromname)
    elif req.has_key('fromname'):
        filter['fromname'] = req.fromname
    else:
        filter['fromname'] = ''

    # email
    if req.has_key('showemail') and \
       req.showemail <> self.when_ignore_word:
            filter['email'] = string.strip(req.showemail)
    elif req.has_key('email'):
        filter['email'] = req.email
    else:
        filter['email'] = ''


    # prepare for caseindependence
    filter = _fixDictofLists(filter)
        
    # all Issue objects
    containers = objsequence 

    #
    # Before we can determine what to sort on
    # we need to check the name of the properties
    
    # the default
    sortorder = ('bobobase_modification_time',)
    skipsort = 0
    if req.has_key('sortorder') and req['sortorder'] != 'date':
        sortorder = (req['sortorder'],)
        if req['sortorder'] == 'from':
            sortorder = ('fromname',)
        if req['sortorder'] == 'creationdate':
            sortorder = ('issuedate',)
        if req['sortorder'] == 'changedate':
            sortorder = ('bobobase_modification_time',)
        if req['sortorder'] == 'status':
            #
            # Very special!
            # self.statuses is an "human ordered" list like 'open','taken',...'completed','forgotten'
            statuses = {}
            # create a dictionary of list divided in groups by their status
            # like this: {'open':['foo','bar'], ... 'forgotten':['this','that']}
            for con in containers:
                if statuses.has_key(con.status):
                    statuses[con.status].append(con)
                else:
                    statuses[con.status] = [con]
            # recreate the containers
            containers = []
            # remembers that self.statuses is ordered
            for status in self.statuses:
                # perhaps there are no 'forgotten' issues even
                # though there is 'forgotten' in self.statuses
                if statuses.has_key(status):
                    # order all issues with this status by bobobase_modification_time
                    thisstatus_containers = sequence.sort(statuses[status],(('bobobase_modification_time',),))
                    # reverse bobobase_modification_time sort naturally
                    thisstatus_containers.reverse()
                    containers = containers + thisstatus_containers
                    # skip the default sort later
                    skipsort = 1
                    
        if req['sortorder'] == 'urgency':
            #
            # Very special too
            # read the comments on "if req['sortorder'] == 'status':" above
            urgencies = {}
            for con in containers:
                if urgencies.has_key(con.urgency):
                    urgencies[con.urgency].append(con)
                else:
                    urgencies[con.urgency] = [con]
            containers = []
            for urgency in self.urgencies+['']:
                if urgencies.has_key(urgency):
                    thisurgency_containers = sequence.sort(urgencies[urgency], (('bobobase_modification_time',),))
                    thisurgency_containers.reverse()
                    containers = containers + thisurgency_containers
                    skipsort = 1
            
    else:
        # if no sortorder,
        # it uses the bobobase_modification_time
        # and this has to be reversed so that
        # the latest is listed first
        reverse =1

    # do the basic sort.
    # 'sortorder' is a tuple
    if not skipsort:
        containers = sequence.sort(containers,(sortorder,))

        
    if reverse:
        containers.reverse()

    # containers is not order, just filter it
    for con in containers:
        # if confidential and (if) Manager show, if not skip
        if not con.confidential or hasManagerRole(self):
            if filter == {}:
                rlist.append(con)
            else:
                passsections = 0
                for section in con.sections:
                    if section.lower() in filter['sections']:
                        passsections = 1
                        break
                # pass the filter lists
                if con.status.lower() in filter['status'] \
                   and con.urgency.lower() in filter['urgency'] \
                   and con.type.lower() in filter['type'] \
                   and passsections:
                    if (con.fromname.lower() == filter['fromname'] or filter['fromname']=='') \
                       and \
                       (con.email.lower() == filter['email'] or filter['email'] =='' ):
                        rlist.append(con)
 
    return rlist

           

def _fixDictofLists(dict):
    " throw it a dictionary and it returns the values lowercased "
    for key,value in dict.items():
        if type(value)==type([]):
            dict[key] = _lowercaseList(value)
        elif type(value)== type('s'):
            dict[key] = value.lower()
    return dict

def _lowercaseList(lst):
    " lowercase and strip all items in a list "
    nlst=[]
    for l in lst:
        nlst.append(l.strip().lower())
    return nlst

def ListThreads(self,issueID):
    """
    returnt the followup objects. The threads.
    """
    req = self.REQUEST
    rlist = []
    issueobject = getattr(self,issueID)

    threads = issueobject.objectValues('Site Tracker Issue Thread')
    for thread in threads:
        # if testing on the thread needs to be done,
        # do that here.
        rlist.append(thread)
        
    return rlist                
                
def ModifyIssue(self,action='Addfollowup'):
    req = self.REQUEST

    # ! Change correspondingly to you pref cookie object !
    cookieobject = self.cookieobject()

    
    action = action.strip().lower()
    oldstatus = self.status

    if not req.has_key('comment_description'):
        req.set('comment_description','')

    req_manager = 1 # require manager
    addfollowup = 0
    past_tense=None
    if action == 'open':
        past_tense='open'
        addfollowup = 1
    if action == 'take':
        past_tense='taken'
        addfollowup = 1
    if action == 'pending':
        past_tense='pending'
        addfollowup = 1
    if action == 'reject':
        past_tense='rejected'
        addfollowup = 1
    if action == 'complete':
        past_tense='completed'
        addfollowup = 1
    if action == 'delete':
        DeleteIssue(self, self.id)
        addfollowup = 0
        redirect_url = req.URL2
    if action == 'addfollowup':
        addfollowup = 1
        req_manager = 0

    if req_manager and not hasManagerRole(self):
        # the chosen action requires manager role,
        # but the person is not a manager
        getattr(self,'login')(self,req)
        
    if past_tense is not None:
        self.propertysheets.properties.manage_changeProperties({'status':past_tense})
    else:
        # anything else means that the comment_description
        # cannot be empty
        if req['comment_description'].strip()=='':
            addfollowup=0
            redirect_url = req.URL1
    
    # most actions may perhaps add a little comment
    if addfollowup:
        genid = RandomID(self, "IssueThread", 5, meta_type='Site Tracker Issue Thread')
        if action == 'addfollowup':
            gentitle = 'Added Issue followup'
        else:
            gentitle = 'Changed status from %s to %s'%\
                       (oldstatus.capitalize(), past_tense.capitalize())

        # create an Thread object
        self.SiteTrackerIssueThread.createInObjectManager(genid,req)

        # get that object
        followupobject = getattr(self, genid)

        # fix variables
        try:
            fromname = req.fromname
        except:
            fromname = ''
        try:
            email = req.email
        except:
            email = ''

            
        # remember visitor
        cookieobject.set(self.name_cookiekey,fromname)
        cookieobject.set(self.email_cookiekey,email)

        # update Thread object
        newproperties = {'title':gentitle,
                         'comment':req.comment_description,
                         'threaddate':getNowdate(),
                         'fromname':fromname,
                         'email':email}
        followupobject.propertysheets.properties.manage_changeProperties(newproperties)

        if req.has_key('notify'):
                    
            # now, create and email-alert-queue object
            # using filtered email address
            # get who to notify        
            email_addresses = Others2Notify(self, genid, do='email',emailtoskip=email)

            if len(email_addresses) > 0:

                # get root instance
                trackerroot = getRoot(self)

                notifyid = RandomID(trackerroot, 'Notification',5,meta_type='Issue Notification')
                trackerroot.IssueNotification.createInObjectManager(notifyid,req)
                
                notifyobject = getattr(trackerroot,notifyid)

                newproperties = {'title':self.title,
                                 'change':gentitle,
                                 'issueID':self.id,
                                 'comment':req.comment_description,
                                 'anchorname':len(self.objectIds('Site Tracker Issue Thread')),
                                 'emails':email_addresses,
                                 'date':getNowdate(),
                                 'fromname':fromname}
                notifyobject.propertysheets.properties.manage_changeProperties(newproperties)
                
                # use the dispatcher to try to send
                # this notification right now.
                req.set('loop_over',[notifyobject])
                # there is no big deal if the dispatcher crashes here
                # because the notification is saved and the dispatcher
                # can be invoked some other time manually
                try:
                    getattr(self,'dispatcher.html')(self,req)
                except: pass 
        
        redirect_url = '%s#%s'%\
                       (self.absolute_url(),len(self.objectIds('Site Tracker Issue Thread')))
        
    # ready !
    # redirect
    req.RESPONSE.redirect(redirect_url)
            


    
def Others2Notify(self, issueID, do='print',emailtoskip=None):
    """
    Returns a list of names and emails of guys to notify on this Issue. 
    Take the hide_me and confidential factor into consideration.
    
    if do == 'print':
        display the people on screen (using hide_me and confidential factor)
    if do == 'email':
        return a pure list of all email address
    """

    cookieobject = self.cookieobject()

    if emailtoskip is None:
        if cookieobject.has_key(self.email_cookiekey):
            emailtoskip = cookieobject[self.email_cookiekey]
        else:
            emailtoskip = '@@@@@'

    fromstrings = []


    issue = self

    # proceed only if Manager or open-name on the issue
    if self.hasManagerRole() or not issue.hide_me:
        # self.email is the email of the thread
        # emailtoskip comes from the param/cookie if applicable
        if not issue.email == emailtoskip:
            if do =='email':
                fromstring = _fromstring(issue.fromname, issue.email, email_must=1, email_pure=1)
            else:
                fromstring = _fromstring(issue.fromname, issue.email)
            # fromstring looks like this now: "Peter Bengtsson <mail@peterbe.com>"
            # or ['mail@peterbe.com'] depending on 'do'
            # ...maybe
            if fromstring and fromstring not in fromstrings:
                fromstrings.append(fromstring)

    for thread in issue.objectValues('Site Tracker Issue Thread'):
        if not thread.email == emailtoskip:
            if do =='email':
                fromstring = _fromstring(thread.fromname, thread.email, email_must=1, email_pure=1)
            else:
                fromstring = _fromstring(thread.fromname, thread.email)
                
            if fromstring and fromstring not in fromstrings:
                
                    fromstrings.append(fromstring)

    return fromstrings


def _filterduplicates(olist):
    nlist = []
    flist = []
    for each in olist:
        if not each[0:1] in nlist:
            nlist.append(each[0:1])
            flist.append(each)
    return flist
            
    
    
    
    
def _fromstring(name, email, email_html=1, email_must=0, email_pure=0):
    """
    First we filter which combs to care about;
    then we filter them.
    """
	
    passit = 0
       
    # first filter
    if name <> '' or email <> '':
        if email_must:
            if ValidEmailAddress(email):
                passit = 1
            else:
                passit = 0
        else:
            passit = 1
    
    # then display
    if not passit:
        return 0
    else:
        if email_pure and ValidEmailAddress(email):
            return email
        elif email == '':
            return name
        else:
            if email_html:
                return string.strip('%s &lt;<a href="mailto:%s">%s</a>&gt;'%(name, email, email))
            else:
                return string.strip('%s <%s>'%(name, email))
    return 0

    
        

                
def hasManagerRole(self):
    """
    This method determines if the current user
    is allowed to do stuff that only the Zope manager is
    supposed to be able to do.
    Feel free to edit appropriatly to what suits you.
    """
    if 'Manager' in self.REQUEST.AUTHENTICATED_USER.getRoles():
        return 1
    # default is false
    return 0


    
    
def HeadingLinks(self, display, sortname, default=0, startreversed=0):
    """Returns a hyperlink that can
       be used for resorting the listing
    """
    req = self.REQUEST
    querystring = req.QUERY_STRING

    ascending = self.ascarrow.tag(hspace=2)
    descending = self.descarrow.tag(hspace=2)
    invisible = self.invisiblearrow.tag(hspace=2)
    
    
    n_querystring = []
    for query in string.split(querystring,'&'):
        if string.find(query,'sortorder') == -1 and string.find(query,'reverse') == -1:
            n_querystring.append(query)
    querystring = string.join(n_querystring,'&')

    if querystring != '':
        querystring = '&%s'%querystring
    
    if req.has_key('reverse') and (req['reverse'] == 1 or req['reverse'] == '1'):
        reverse = 1
    elif startreversed and not (req.has_key('sortorder') and req['sortorder']==sortname):
        reverse = 1
    else:
        reverse = 0
    if req.has_key('sortorder') and req['sortorder'] == sortname or (default and not req.has_key('sortorder')):
        # have sorted by this, just let them reverse
        if reverse:
            return '%s<a class="headinglink_current" href="%s?sortorder=%s%s" title="Click to sort by %s">%s</a>' \
                   % (descending,req.URL, sortname, querystring, display, display )
        else:
            return '%s<a class="headinglink_current" href="%s?sortorder=%s&reverse=1%s" title="Click one more time and the sorting will be reversed">%s</a>'\
                   % (ascending,req.URL, sortname, querystring, display )
    else:
        if startreversed:
            return '%s<a class="headinglink" href="%s?sortorder=%s%s&reverse=1" title="Click to sort by %s">%s</a>' \
               % (invisible,req.URL, sortname, querystring, display, display )
        else:
            return '%s<a class="headinglink" href="%s?sortorder=%s%s" title="Click to sort by %s">%s</a>' \
               % (invisible,req.URL, sortname, querystring, display, display )
        
def MarkNotificationDispatch(self):
    """ set the 'dispatch' unrestricted of security """
    self.propertysheets.properties.manage_changeProperties({'dispatched':1})
    
